/*

*/

package language6;

import tree.*;
import tokenizer.*;

/**
   This program takes an abstract syntax tree (AST)
   that is defined by the following grammar and
   converts it into a string that contains a
   C/C++/Java/JavaScript like infix notation.
   <pre>{@code
       Prog ::= Expr
              | '(' 'prog' Expr* Expr ')'

       Expr ::= Lambda
              | Apply
              | If
              | While
              | Begin
              | Var
              | Set
              | Print
              | AExpr
              | BExpr
              | INTEGER
              | BOOLEAN
              | VARIABLE

     Lambda ::= '(' 'lambda' VARIABLE* Expr ')'  // formal parameters followed by function body

      Apply ::= '(' 'apply' Expr Expr* ')'  // function value followed by actual parameters

         If ::= '(' 'if' Expr Expr Expr ')'

      While ::= '(' 'while' Expr Expr ')'

      Begin ::= '(' 'begin' Expr+ ')'

        Var ::= '(' 'var' VARIABLE Expr ')'

        Set ::= '(' 'set' VARIABLE Expr ')'

      Print ::= '(' 'print' Expr ')'

      BExpr ::= '(' '||'  Expr Expr+ ')'
              | '(' '&&'  Expr Expr+ ')'
              | '(' '!'   Expr ')'
              | '('  EqOp Expr Expr ')'
              | '(' RelOp Expr Expr ')'

       EqOp ::= '==' | '!='
      RelOp ::= '<' | '>' | '<=' | '>='

      AExpr ::= '(' '+' Expr Expr* ')'
              | '(' '-' Expr Expr? ')'
              | '(' '*' Expr Expr+ ')'
              | '(' '/' Expr Expr  ')'
              | '(' '%' Expr Expr  ')'
              | '(' '**' Expr Expr ')'

    INTEGER ::= [-|+][0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
   }</pre>
*/

public class AST2infix
{
   public static String ast2infix(Tree<Token> tree)
   {
      return convertProg( tree, "", true );
   }


   // Convert a prog
   public static String convertProg(final Tree<Token> tree,
                                    final String indent,
                                    final boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      // Check which kind of Prog we have.
      if ( ! tree.root().lexeme.equals("prog") )
      {
         // Convert the single expression.
         result += convertExp( tree, "", false );
      }
      else
      {
         result += "Program\n{\n";

         // convert each expression
         for (int i = 0; i < tree.subTrees().size(); ++i)
         {
            result += convertExp( tree.subTrees().get(i), "   ", false );

            if ( tree.subTrees().get(i).root().lexeme.equals("fun") )
            {  // give function definitions an extra \n
               result += "\n";
            }

            if ( tree.subTrees().get(i).root().lexeme.equals("begin")
              || tree.subTrees().get(i).root().lexeme.equals("if")
              || tree.subTrees().get(i).root().lexeme.equals("while") )
            {  // the above expressions do not need a semicolon
               result += "\n";
            }
            else // include a semicolon
            {
               result += ";\n";
            }
         }

         result += "}";
      }

      return result;
   }


   // Convert an expression
   public static String convertExp(final Tree<Token> tree,
                                   final String indent,
                                   final boolean inline)
   {
      String result = "";

      String node = tree.root().lexeme;

      if ( node.equals("lambda") )
         result += convertLambda( tree, indent, inline );
      else if ( node.equals("apply") )
         result += convertApply( tree, indent, inline );
      else if ( node.equals("if") )
         result += convertIf( tree, indent, inline );
      else if ( node.equals("while") )
         result += convertWhile( tree, indent, inline );
      else if ( node.equals("begin") )
         result += convertBegin( tree, indent, inline );
      else if ( node.equals("var") )
         result += convertVar( tree, indent, inline );
      else if ( node.equals("set") )
         result += convertSet( tree, indent, inline );
      else if ( node.equals("print") )
         result += convertPrint( tree, indent, inline );
      else if ( node.equals("&&") || node.equals("||")
             || node.equals("!")
             || node.equals("<")  || node.equals(">")
             || node.equals("<=") || node.equals(">=")
             || node.equals("==") || node.equals("!=")
             || node.equals("+")  || node.equals("-")
             || node.equals("*")  || node.equals("/")
             || node.equals("%")  || node.equals("^") )
         result += convertBRAexp( tree, indent, inline );
      else if ( tree.subTrees().size() == 0 )
      {
         result += inline ? node : indent + node;
      }
      else // shouldn't get here; just convert to an S-expression
      {
         result += inline ? tree : indent + tree;
      }

      return result;
   }


   // Convert an anonymous function (a lambda expression)
   public static String convertLambda(final Tree<Token> tree,
                                      final String indent,
                                      final boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      // get the number of formal parameters from the lambda expression
      int numberOfParameters = tree.subTrees().size() - 1;
      // do the list of formal parameters from the lambda expression
      if ( 0 == numberOfParameters )
      {
         result += "()";
      }
      else if ( 1 == numberOfParameters )
      {
         // get the formal parameter
         result += tree.subTrees().get(0).root().lexeme;
      }
      else
      {
         result += "(";
         // get the first formal parameter
         result += tree.subTrees().get(0).root().lexeme;
         for (int i = 1; i < numberOfParameters; ++i)
         {
            String variable = tree.subTrees().get(i).root().lexeme;
            result += ", " + variable;
         }
         result += ")";
      }

      // use a Maple like notation
      result += " -> ";

      // get the body of the function
      result += convertExp( tree.subTrees().get(numberOfParameters), indent, true );

      return result;
   }


   // Convert a function call
   public static String convertApply(final Tree<Token> tree,
                                     final String indent,
                                     final boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      // do the function value
      if ( tree.subTrees().get(0).degree() > 0 )  // if the function value is more
         result += "(";                           // complicated than just a name

      result += convertExp( tree.subTrees().get(0), indent, true);

      if ( tree.subTrees().get(0).degree() > 0 )
      {
         result += ")";
      }

      // do the actual parameters
      result += "(";
      for (int i = 1; i < tree.subTrees().size(); ++i)
      {
         result += convertExp( tree.subTrees().get(i), indent, true );
         if (i < tree.subTrees().size() - 1)
         {
            result += ", ";
         }
      }
      result += ")";

      return result;
   }


   // Convert an if-expression
   public static String convertIf(final Tree<Token> tree,
                                  final String indent,
                                  final boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      result += "if (" + convertExp( tree.subTrees().get(0), indent+"       ", true ) + ")\n";

      if ( tree.subTrees().get(1).root().lexeme.equals("begin") )
      {
         result += convertExp( tree.subTrees().get(1), indent, false );
      }
      else if ( tree.subTrees().get(1).root().lexeme.equals("begin")
             || tree.subTrees().get(1).root().lexeme.equals("if")
             || tree.subTrees().get(1).root().lexeme.equals("while") )
      {
         result += convertExp( tree.subTrees().get(1), indent+"   ", false );
      }
      else
      {
         result += convertExp( tree.subTrees().get(1), indent+"   ", false );
         result += ";";
      }

      result += "\n" + indent + "else\n";

      if ( tree.subTrees().get(2).root().lexeme.equals("begin") )
      {
         result += convertExp( tree.subTrees().get(2), indent, false );
      }
      else if ( tree.subTrees().get(2).root().lexeme.equals("begin")
             || tree.subTrees().get(2).root().lexeme.equals("if")
             || tree.subTrees().get(2).root().lexeme.equals("while") )
      {
         result += convertExp( tree.subTrees().get(2), indent+"   ", false );
      }
      else
      {
         result += convertExp( tree.subTrees().get(2), indent+"   ", false ) + ";";
      }

      return result;
   }


   // Convert a while-loop expression
   public static String convertWhile(final Tree<Token> tree,
                                     final String indent,
                                     final boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      result += "while (" + convertExp(tree.subTrees().get(0), indent+"       ", true) + ")\n";

      if ( tree.subTrees().get(1).root().lexeme.equals("begin") )
      {
         result += convertExp( tree.subTrees().get(1), indent, false );
      }
      else if ( tree.subTrees().get(1).root().lexeme.equals("begin")
             || tree.subTrees().get(1).root().lexeme.equals("if")
             || tree.subTrees().get(1).root().lexeme.equals("while") )
      {
         result += convertExp( tree.subTrees().get(1), indent+"   ", false );
      }
      else
      {
         result += convertExp( tree.subTrees().get(1), indent+"   ", false );
         result += ";";
      }

      return result;
   }


   // Convert a begin expression
   public static String convertBegin(final Tree<Token> tree,
                                     final String indent,
                                     final boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      result += "{\n";

      // process each expression
      for(int i = 0; i < tree.subTrees().size(); ++i)
      {
         result += convertExp( tree.subTrees().get(i), indent + "   ", false );
         if ( tree.subTrees().get(i).root().lexeme.equals("begin")
           || tree.subTrees().get(i).root().lexeme.equals("if")
           || tree.subTrees().get(i).root().lexeme.equals("while") )
         {  // the above expressions do not need a semicolon
            result += "\n";
         }
         else
         {
            result += ";\n";
         }
      }

      result += indent + "}";

      return result;
   }


   // Convert a var expression
   public static String convertVar(final Tree<Token> tree,
                                   final String indent,
                                   final boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      // convert the left-hand-side of the declaration expression
      result += "var " + convertExp( tree.subTrees().get(0), indent, true );

      result += " = ";

      // convert the right-hand-side of the declaration expression
    //indentation += 7 + tree.subTrees().get(0).root().lexeme.length();
      result += convertExp( tree.subTrees().get(1), "     "+indent, true );

      return result;
   }


   // Convert a set expression
   public static String convertSet(final Tree<Token> tree,
                                   final String indent,
                                   final boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      // convert the left-hand-side of the assignment expression
      result += convertExp( tree.subTrees().get(0), indent, true );

      result += " = ";

      // convert the right-hand-side of the assignment expression
    //indentation += 3 + tree.subTrees().get(0).root().lexeme.length();
      result += convertExp( tree.subTrees().get(1), "   "+indent, true );

      return result;
   }//convertSet()


   // Convert a print expression
   public static String convertPrint(final Tree<Token> tree,
                                     final String indent,
                                     final boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      result += "print(" + convertExp( tree.subTrees().get(0), indent, true ) + ")";

      return result;
   }


   // Convert a boolean, relational, or arithmetic expression
   public static String convertBRAexp(final Tree<Token> tree,
                                      final String indent,
                                      final boolean inline)
   {
      String result = "";

      if ( ! inline )
         result += indent;

      if ( tree.subTrees().size() == 1 )  // unary operator
      {
         if ( tree.subTrees().get(0).degree() >= 2 )
         {
            result += tree.root().lexeme + "(" + convertExp( tree.subTrees().get(0), indent, true ) + ")";
         }
         else
         {
            result += tree.root().lexeme + convertExp( tree.subTrees().get(0), indent, true );
         }
      }
      else if ( tree.subTrees().size() == 2 )  // binary operator
      {
         int middlePrecedence = precedence( tree.root().lexeme );
         int leftPrecedence = precedence( tree.subTrees().get(0).root().lexeme );

         if ( leftPrecedence > middlePrecedence )
         {
            result += "(" + convertExp( tree.subTrees().get(0), indent, true ) + ")";
         }
         else
         {
            result += convertExp( tree.subTrees().get(0), indent, true );
         }

         result += " " + tree.root().lexeme + " ";

         int rightPrecedence = precedence( tree.subTrees().get(1).root().lexeme );

         if ( rightPrecedence >= middlePrecedence )
         {
            result += "(" + convertExp( tree.subTrees().get(1), indent, true ) + ")";
         }
         else
         {
            result += convertExp( tree.subTrees().get(1), indent, true );
         }
      }
      else  // operators with 3 or more operands
      {
         int operatorPrecedence = precedence( tree.root().lexeme );
         int leftPrecedence = precedence( tree.subTrees().get(0).root().lexeme );

         if ( leftPrecedence > operatorPrecedence )
         {
            result += "(" + convertExp( tree.subTrees().get(0), indent, true ) + ")";
         }
         else
         {
            result += convertExp( tree.subTrees().get(0), indent, true );
         }

         for (int i = 1; i < tree.subTrees().size(); ++i)
         {
            result += " " + tree.root().lexeme + " ";

            int rightPrecedence = precedence( tree.subTrees().get(i).root().lexeme );

            if ( rightPrecedence >= operatorPrecedence )
            {
               result += "(" + convertExp( tree.subTrees().get(i), indent, true ) + ")";
            }
            else
            {
               result += convertExp( tree.subTrees().get(i), indent, true );
            }
         }
      }
      return result;
   }



   // Add the 'var', 'set' and 'print' operators to this function.
   // http://www.java-tips.org/java-se-tips/java.lang/what-is-java-operator-precedence.html
   // http://www.cs.princeton.edu/introcs/11precedence/
   private static int precedence(final String op)
   {
      int result = 0;  // "highest" precedence (i.e., constants)

      if ( op.equals("!") )
      {
         result = 2;
      }
      else if ( op.equals("**") )
      {
         result = 3;
      }
      else if ( op.equals("*")
             || op.equals("/")
             || op.equals("%") )
      {
         result = 4;
      }
      else if ( op.equals("+")
             || op.equals("-") )
      {
         result = 5;
      }
      else if ( op.equals("<")
             || op.equals(">")
             || op.equals("<=")
             || op.equals(">=") )
      {
         result = 7;
      }
      else if ( op.equals("==")
             || op.equals("!=") )
      {
         result = 8;
      }
      else if ( op.equals("&&") )
      {
         result = 11;
      }
      else if ( op.equals("||") )
      {
         result = 12;
      }
      else if ( op.equals("var") )
      {
         result = 14;
      }
      else if ( op.equals("set") )
      {
         result = 14;
      }
      else if ( op.equals("print") )
      {
         result = 14;
      }

      return result;
   }
}
