/*

*/

package language6;

import java.util.ArrayList;

/**
   An {@code Environment} object holds (variable, value) pairs.
   <p>
   The {@code Environment} interface has methods for
   <ol>
   <li> Adding a new (variable, value) pair to this {@code Environment} object.</li>
   <li> Looking up a variable to see if it is in the Environment chain.</li>
   <li> Looking up a variable to see if it is in this {@code Environment} object.</li>
   <li> Looking up a variable's value from the variable's (variable, value) pair.</li>
   <li> Mutating the value part of a variable's (variable, value) pair.</li>
   </ol>
*/
public class Environment
{
   private final ArrayList<String> variables;
   private final ArrayList<Value>  values;
   private final Environment nonLocalLink; // used to look up "non-local" variables
   private final String label;              // used in debugging output

   /**
      Create an empty "global" {@code Environment} object.
   */
   public Environment()
   {
      this(null, "Global"); // denotes the global environment object
   }


   /**
      Create an empty "nested" {@code Environment} object.

      @param env   {@code Environment} object that this {@code Environment} object is nested within
      @param label  used in debugging output
   */
   public Environment(Environment env, String label)
   {
      variables = new ArrayList<String>();
      values    = new ArrayList<Value>();
      nonLocalLink = env;
      this.label = label;
   }


   /**
      Look up {@code variable} in the environment chain.
      Return true if {@code variable} is in the environment
      chain, otherwise return false.

      @param variable  to be looked up in the environment chain
      @return true if the variable is found in the environment chain
   */
   public boolean defined(String variable)
   {
      return ( null != this.lookUp(variable) );
   }


   /**
      Look up {@code variable} in this {@code Environment} object.
      Return true if {@code variable} is in this {@code Environment}
      object, otherwise return false.

      @param variable  to be looked up in this {@code Environment} object
      @return true if the variable is found in this {@code Environment} object
   */
   public boolean definedLocal(String variable)
   {
      int i;
      for (i = 0; i < variables.size(); ++i)
         if ( variable.equals(variables.get(i)) )
            break;

      return ( i < variables.size() );
   }


   /**
      Recursively look up {@code variable} in the environment
      chain and return its associated {@link Value}.
      <p>
      Return null if {@code variable} is not found in the
      environment chain.

      @param variable  to be looked up in the environment chain
      @return the {@link Value} of the variable, if it is found in the environment chain
   */
   public Value lookUp(String variable)
   {
      int i;
      for (i = 0; i < variables.size(); ++i)
         if ( variable.equals(variables.get(i)) )
            break;

      if ( i < variables.size() )
      {
         return values.get(i);
      }
      else
      {
         if ( null == nonLocalLink )
            return null; // variable was not found
         else // recursively search the rest of the environment chain
            return nonLocalLink.lookUp(variable); // recursion
      }
   }


   /**
      Add a (variable, value) pair to this {@code Environment} object.

      @param variable to be added to this {@code Environment}
      @param value {@link Value} object to associate with the given variable
   */
   public void add(String variable, Value value)
   {
      variables.add(variable);
         values.add(value);
   }


   /**
      Update the {@link Value} associated with {@code variable}
      in the environment chain. Return true if the update
      is successful, return false if {@code variable} is not found
      in the environment chain.

      @param variable  to be updated in the environment chain
      @param value  {@link Value} to put in the environment chain for the given variable
      @return true if the variable was found and updated
   */
   public boolean update(String variable, Value value)
   {
      int i;
      for (i = 0; i < variables.size(); ++i)
         if ( variable.equals(variables.get(i)) )
            break;

      if ( i < variables.size() )
      {
         values.set(i, value);
         return true;
      }
      else
      {
         if ( null == nonLocalLink )
            return false; // variable cannot be found
         else // recursively search the rest of the environment chain
            return nonLocalLink.update(variable, value); // recursion
      }
   }


   /**
      Convert the contents of the environment chain into a {@link String}.
      This is mainly for debugging purposes.

      @return a {@link String} representation of the entire environment chain
   */
   @Override
   public String toString()
   {
      String result = "";

      // Convert this Environment object first.
      result += "[[ " + label + " Environment";
      for (int i = 0; i < variables.size(); ++i)
      {
         result += "\n[ " + variables.get(i) + " = " + values.get(i);
      }
      result += "\n]]";

      // Now recursively convert the environment chain.
      if ( null == nonLocalLink )
         return result;
      else
         return result + "\n||\n\\/\n" + nonLocalLink.toString();
   }
/*
   public String toString()
   {
      String result = "";

      // Recursively convert the environment chain first.
      if ( null != nonLocalLink )
         result = nonLocalLink.toString() + "\n/\\\n||\n[[ " + label + " Environment";
      else
         result += "[[" + label + " Environment";

       // Now convert this Environment object.
      for (int i = 0; i < variables.size(); ++i)
      {
         result += "\n[ " + variables.get(i) + " = " + values.get(i);
      }

      return result;
   }
*/
}
