/*

*/

package language6;

import tree.Tree;
import tree.LinkedTree;
import tokenizer.Token;

/**
   This class is a kind of "tagged union".
   <p>
   Each instance of this class holds a value for our simple language.
   But our language has three kinds of values, ints, booleans, and
   functions (closures). So each instance of this class holds both
   an int value, a boolean value, and a closure (IPEP) value. The
   "tag" field is used to determine which of the three values that
   an instance holds, the int or the boolean or the closure, is valid.
   <p>
   See <a href="https://en.wikipedia.org/wiki/Tagged_union" target="_top">
                https://en.wikipedia.org/wiki/Tagged_union</a>
*/
public class Value
{
   public final String tag;     // either "int" or "bool" or "lambda"
   public final int valueI;
   public final boolean valueB;
   public final IPEP valueL;    // change valueL from Tree to IPEP

   public static final String INT_TAG = "int";
   public static final String BOOL_TAG = "bool";
   public static final String CLOSURE_TAG = "closure";

   /**
      Construct a {@code Value} object that holds an int value.

      @param value  the integer value for this {@code Value} object
   */
   public Value(int value)
   {
      this.tag = INT_TAG;
      this.valueI = value;
      // Implement C/C++ like semantics.
      this.valueB = (value != 0);
      valueL = new IPEP(new LinkedTree<Token>(new Token("null",0,0)),
                        null); // empty closure
   }


   /**
      Construct a {@code Value} object that holds a boolean value.

      @param value  the boolean value for this {@code Value} object
   */
   public Value(boolean value)
   {
      this.tag = BOOL_TAG;
      this.valueB = value;
      // Implement C/C++ like semantics.
      this.valueI = (value ? 1 : 0);
      valueL = new IPEP(new LinkedTree<Token>(new Token("null",0,0)),
                        null); // empty closure
   }


   /**
      Construct a {@code Value} object that holds a "closure".

      @param value  the closure object for this {@code Value} object
   */
   public Value(IPEP value)
   {
      this.tag = CLOSURE_TAG;
      this.valueL = value;
      this.valueB = true;
      this.valueI = 1;
   }


   @Override
   public String toString()
   {
      return "[tag->" + tag
                      + ", valueI->" + valueI
                      + ", valueB->" + valueB
                      + ", valueL->" + valueL
                      + "]";
   }


   /**
      @return a simplified {@link String} representation of this {@code Value}
   */
   public String toSimpleString()
   {
      String result = "";

      if ( tag.equals(BOOL_TAG) )
      {
         result += valueB;
      }
      else if ( tag.equals(INT_TAG) )
      {
         result += valueI;
      }
      else if ( tag.equals(CLOSURE_TAG) )
      {
         result += valueL;
      }
      else // bad tag (shouldn't get here)
      {
         result += "[tag->" + tag
                            + ", valueI->" + valueI
                            + ", valueB->" + valueB
                            + ", valueL->" + valueL
                            + "]";
      }

      return result;
   }
}
