/*

*/

package treelanguage;

import tree.*;
import tokenizer.*;

/**
   Parse a {@link String} in the Tree language into a
   {@link Tree} of {@link String}s.
   <p>
   A {@link String} in the Tree language is defined by
   the following grammar.
   <pre>{@code
     Tree ::= '(' String Tree+ ')'
            | String
   }</pre>
   where {@code '('} and {@code ')'} are literals,
   and {@code Tree+} means "one or more".
   <p>
   Input strings are tokenized by {@link Tokenizer}.
   <p>
   The static method {@link #getTree} is essentially a
   recursive descent parser for this Tree language.
*/
public class Parse2AST
{
   /**
      Parse a {@link String} from the Tree language into
      an abstract syntax {@link Tree} of {@link Token}s.

      @param expression  {@link String} from the Tree language to parse into a {@link Tree}
      @return the {@link Tree} represented by {@code expression}
      @throws TokenizeError if {@code expression} fails to tokenize
      @throws ParseError if {@code expression} fails to parse
   */
   public static Tree<Token> parse(final String expression)
   {
      final Tokenizer tokens = new Tokenizer(expression);
      final Tree<Token> tree = parse(tokens);
      return tree;
   }


   /**
      Parse a {@link Token} sequence representing a string
      from the Tree language. Return the instance of the
      {@link Tree} class that is represented by the
      {@link Token} sequence.

      @param tokens  {@link java.util.List} of {@link Token}s to parse into a {@link Tree}
      @return the {@link Tree} represented by the {@link Token} sequence
      @throws ParseError if the {@link Token} sequence fails to parse
   */
   public static Tree<Token> parse(final Tokenizer tokens)
   {
      final Tree<Token> tree = getTree(tokens); // Parse the tokens.

      if ( tokens.hasToken() ) // There shouldn't be any more tokens.
         throw new ParseError("unexpected input: "
                            + tokens.peek().lexeme
                            + " at line " + tokens.peek().line
                            + ", position " + tokens.peek().position
                             );
      return tree;
   }


   /**
      This private static method is essentially a
      recursive descent parser for the Tree
      production in the Tree grammar.

      @param tokens  {@link java.util.List} of {@link Token}s to parse into a {@link Tree}
      @return the {@link Tree} of {@link Token}s that the tokens represent
      @throws ParseError if the {@link Token} sequence fails to parse
   */
   private static Tree<Token> getTree(final Tokenizer tokens)
   {
      if ( tokens.isEmpty() ) // There should be another token.
         throw new ParseError("unexpected end of input");

      final Token token = tokens.next(); // Consume one token.

      if ( token.lexeme.equals("(") ) // Look for a parenthesized tree.
      {
         if ( tokens.isEmpty() ) // There should be another token.
            throw new ParseError("unexpected end of input");

         // Consume the root of the tree.
         final Tree<Token> result = new LinkedTree<>(tokens.next());

         if ( tokens.isEmpty() ) // There should be another token.
            throw new ParseError("unexpected end of input");

         do
         {  // Consume the next sub tree.
            result.subTrees().add( getTree(tokens) );

            if ( tokens.isEmpty() ) // There should be another token.
               throw new ParseError("unexpected end of input");
         }
         while (! tokens.peek().lexeme.equals(")")); // One token look ahead.

         tokens.match(")"); // Consume the matching ")".

         return result;
      }
      else // The tree must be just the root.
      {
         return new LinkedTree<>(token);
      }
   }
}
