/*

*/

package treelanguage;

import tree.*;
import tokenizer.*;

/**
   Parse a {@link String} in the Tree language into a
   parse {@link Tree} of {@link String}s.
   <p>
   A {@link String} in the Tree language is defined by
   the following grammar.
   <pre>{@code
     Tree ::= '(' String Tree+ ')'
            | String
   }</pre>
   where {@code '('} and {@code ')'} are literals,
   and {@code Tree+} means "one or more".
   <p>
   Input strings are tokenized by {@link Tokenizer}.
   <p>
   The static method {@link #getTree} is essentially a
   recursive descent parser for this Tree language.
*/
public class Parse2SyntaxTree
{
   /**
      Parse a {@link String} from the Tree language into
      a parse {@link Tree} of {@link String}s.

      @param expression  {@link String} from the Tree language to parse
      @return the parse {@link Tree} of the string
      @throws TokenizeError if {@code expression} fails to tokenize
      @throws ParseError if {@code expression} fails to parse
   */
   public static Tree<String> parse(final String expression)
   {
      final Tokenizer tokens = new Tokenizer(expression);
      final Tree<String> tree = parse(tokens);
      return tree;
   }


   /**
      Parse a {@link Token} sequence representing a string
      from the Tree language. Return the parse {@link Tree}
      of the string represented by the {@link Token}s.

      @param tokens  {@link java.util.List} of {@link Token}s to parse into a parse {@link Tree}
      @return the parse {@link Tree} of the tokens
      @throws ParseError if the {@link Token} sequence fails to parse
   */
   public static Tree<String> parse(final Tokenizer tokens)
   {
      final Tree<String> tree = getTree(tokens); // parse the tokens

      if ( tokens.hasToken() ) // there shouldn't be any more tokens
         throw new ParseError("unexpected input: "
                            + tokens.peek().lexeme
                            + " at line " + tokens.peek().line
                            + ", position " + tokens.peek().position
                             );
      return tree;
   }


   /**
      This private static method is essentially a
      recursive descent parser for the Tree
      production in the Tree grammar.

      @param tokens  {@link java.util.List} of {@link Token}s to parse into a parse {@link Tree}
      @return the parse {@link Tree} of the tokens
      @throws ParseError if the {@link Token} sequence fails to parse
   */
   private static Tree<String> getTree(final Tokenizer tokens)
   {
      final Tree<String> result = new LinkedTree<>("Tree");

      if ( tokens.isEmpty() ) // There should be another token.
         throw new ParseError("unexpected end of input");

      final Token token = tokens.next(); // Consume one token.

      if ( token.lexeme.equals("(") ) // Look for a parenthesized tree.
      {
         result.subTrees().add( new LinkedTree<>(token.toString()) ); // add the '('

         if ( tokens.isEmpty() ) // There should be another token.
            throw new ParseError("unexpected end of input");

         // Consume the root of the tree.
         result.subTrees().add( new LinkedTree<>(tokens.next().toString()) );

         if ( tokens.isEmpty() ) // There should be another token.
            throw new ParseError("unexpected end of input");

         do
         {  // Consume the next sub tree.
            result.subTrees().add( getTree(tokens) );

            if ( tokens.isEmpty() ) // There should be another token.
               throw new ParseError("unexpected end of input");
         }
         while (! tokens.peek().lexeme.equals(")")); // One token look ahead.

         final Token token2 = tokens.next(); // Consume the matching ")".

         result.subTrees().add( new LinkedTree<>(token2.toString()) ); // add the ')'
      }
      else // The tree must be just the root.
      {
         result.subTrees().add( new LinkedTree<>(token.toString()) );
      }
      return result;
   }
}
